@@ -2,7 +2,7 @@
  * ProFTPD: mod_ifsession -- a module supporting conditional
  *                            per-user/group/class configuration contexts.
  *
- * Copyright (c) 2002-2008 TJ Saunders
+ * Copyright (c) 2002-2011 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders and other respective copyright holders
  * give permission to link this program with OpenSSL, and distribute the
@@ -26,16 +26,16 @@
  * This is mod_ifsession, contrib software for proftpd 1.2 and above.
  * For more information contact TJ Saunders <tj@castaglia.org>.
  *
- * $Id: mod_ifsession.c,v 1.24 2008/10/14 22:28:02 castaglia Exp $
+ * $Id: mod_ifsession.c,v 1.37 2011/05/23 20:56:40 castaglia Exp $
  */
 
 #include "conf.h"
 
-#define MOD_IFSESSION_VERSION	"mod_ifsession/1.0"
+#define MOD_IFSESSION_VERSION	"mod_ifsession/1.1"
 
 /* Make sure the version of proftpd is as necessary. */
-#if PROFTPD_VERSION_NUMBER < 0x0001021001
-# error "ProFTPD 1.2.10rc1 or later required"
+#if PROFTPD_VERSION_NUMBER < 0x0001030402
+# error "ProFTPD 1.3.4rc2 or later required"
 #endif
 
 #define IFSESS_CLASS_NUMBER	100
@@ -45,14 +45,21 @@
 #define IFSESS_USER_NUMBER	102
 #define	IFSESS_USER_TEXT	"<IfUser>"
 
+module ifsession_module;
+
+static int ifsess_ctx = -1;
 static int ifsess_merged = FALSE;
 
+static const char *trace_channel = "ifsession";
+
 /* Support routines
  */
 
 static void ifsess_remove_param(xaset_t *set, const char *name) {
   config_rec *c = NULL;
 
+  pr_trace_msg(trace_channel, 9, "removing '%s' config", name);
+
   c = find_config(set, -1, name, TRUE);
   while (c != NULL) {
     xaset_t *fset;
@@ -70,6 +77,8 @@ static void ifsess_dup_param(pool *dst_pool, xaset_t **dst, config_rec *c,
     config_rec *parent) {
   config_rec *dup_c = NULL;
 
+  pr_trace_msg(trace_channel, 9, "adding '%s' config", c->name);
+
   if (!*dst)
     *dst = xaset_create(dst_pool, NULL);
 
@@ -104,8 +113,12 @@ static void ifsess_dup_param(pool *dst_pool, xaset_t **dst, config_rec *c,
        * directive then effectively replaces any directive there.
        */
       if (c->config_type == CONF_PARAM &&
-          !(c->flags & CF_MERGEDOWN_MULTI))
+          !(c->flags & CF_MERGEDOWN_MULTI) &&
+          !(c->flags & CF_MULTI)) {
+          pr_trace_msg(trace_channel, 15, "removing '%s' config because "
+            "c->flags does not contain MULTI or MERGEDOWN_MULTI", c->name);
         ifsess_remove_param(dup_c->subset, c->name);
+      }
 
       ifsess_dup_param(dst_pool, &dup_c->subset, c, dup_c);
     }
@@ -129,8 +142,11 @@ static void ifsess_dup_set(pool *dst_pool, xaset_t *dst, xaset_t *src) {
      * directive then effectively replaces any directive there.
      */
     if (c->config_type == CONF_PARAM &&
-        !(c->flags & CF_MERGEDOWN_MULTI))
+        !(c->flags & CF_MERGEDOWN_MULTI)) {
+        pr_trace_msg(trace_channel, 15, "removing '%s' config because "
+          "c->flags does not contain MERGEDOWN_MULTI", c->name);
       ifsess_remove_param(dst, c->name);
+    }
 
     ifsess_dup_param(dst_pool, &dst, c, NULL);
   }
@@ -161,17 +177,17 @@ MODRET start_ifctxt(cmd_rec *cmd) {
 
   if (strcmp(cmd->argv[0], IFSESS_CLASS_TEXT) == 0) {
     name = "_IfClassList";
-    config_type = IFSESS_CLASS_NUMBER;
+    ifsess_ctx = config_type = IFSESS_CLASS_NUMBER;
     eval_type = PR_EXPR_EVAL_OR;
 
   } else if (strcmp(cmd->argv[0], IFSESS_GROUP_TEXT) == 0) {
     name = "_IfGroupList";
-    config_type = IFSESS_GROUP_NUMBER;
+    ifsess_ctx = config_type = IFSESS_GROUP_NUMBER;
     eval_type = PR_EXPR_EVAL_AND;
 
   } else if (strcmp(cmd->argv[0], IFSESS_USER_TEXT) == 0) {
     name = "_IfUserList";
-    config_type = IFSESS_USER_NUMBER;
+    ifsess_ctx = config_type = IFSESS_USER_NUMBER;
     eval_type = PR_EXPR_EVAL_OR;
   }
 
@@ -179,52 +195,54 @@ MODRET start_ifctxt(cmd_rec *cmd) {
    * regular expression?
    */
   if (cmd->argc-1 > 1) {
-    if (strcmp(cmd->argv[1], "AND") == 0) {
+    if (strncmp(cmd->argv[1], "AND", 4) == 0) {
       eval_type = PR_EXPR_EVAL_AND;
       argc = cmd->argc-2;
       argv = cmd->argv+1;
 
-    } else if (strcmp(cmd->argv[1], "OR") == 0) {
+    } else if (strncmp(cmd->argv[1], "OR", 3) == 0) {
       eval_type = PR_EXPR_EVAL_OR;
       argc = cmd->argc-2;
       argv = cmd->argv+1;
 
-    } else if (strcmp(cmd->argv[1], "regex") == 0) {
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-      regex_t *preg = NULL;
+    } else if (strncmp(cmd->argv[1], "regex", 6) == 0) {
+#ifdef PR_USE_REGEX
+      pr_regex_t *pre = NULL;
       int res = 0;
 
       if (cmd->argc != 3)
         CONF_ERROR(cmd, "wrong number of parameters");
 
-      preg = pr_regexp_alloc();
+      pre = pr_regexp_alloc(&ifsession_module);
 
-      res = regcomp(preg, cmd->argv[2], REG_EXTENDED|REG_NOSUB);
+      res = pr_regexp_compile(pre, cmd->argv[2], REG_EXTENDED|REG_NOSUB);
       if (res != 0) {
         char errstr[200] = {'\0'};
 
-        regerror(res, preg, errstr, sizeof(errstr));
-        pr_regexp_free(preg);
+        pr_regexp_error(res, pre, errstr, sizeof(errstr));
+        pr_regexp_free(NULL, pre);
 
         CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": '", cmd->argv[2], "' failed "
           "regex compilation: ", errstr, NULL));
       }
 
-      c = add_config_param(name, 2, NULL, NULL);
+      eval_type = PR_EXPR_EVAL_REGEX;
+
+      c = add_config_param(name, 3, NULL, NULL, NULL);
       c->config_type = config_type;
-      c->argv[0] = pcalloc(c->pool, sizeof(unsigned char));
-      *((unsigned char *) c->argv[0]) = PR_EXPR_EVAL_REGEX;
-      c->argv[1] = (void *) preg;
+      c->argv[0] = pstrdup(c->pool, cmd->arg);
+      c->argv[1] = pcalloc(c->pool, sizeof(unsigned char));
+      *((unsigned char *) c->argv[1]) = eval_type;
+      c->argv[2] = (void *) pre;
 
       return PR_HANDLED(cmd);
 
 #else
       CONF_ERROR(cmd, "The 'regex' parameter cannot be used on this system, "
         "as you do not have POSIX compliant regex support");
-#endif /* HAVE_REGEX_H && HAVE_REGCOMP */
+#endif /* regex support */
 
     } else {
-
       argc = cmd->argc-1;
       argv = cmd->argv;
     }
@@ -239,12 +257,13 @@ MODRET start_ifctxt(cmd_rec *cmd) {
   c = add_config_param(name, 0);
 
   c->config_type = config_type;
-  c->argc = acl->nelts + 1;
-  c->argv = pcalloc(c->pool, (c->argc + 1) * sizeof(char *));
-  c->argv[0] = pcalloc(c->pool, sizeof(unsigned char));
-  *((unsigned char *) c->argv[0]) = eval_type;
+  c->argc = acl->nelts + 2;
+  c->argv = pcalloc(c->pool, (c->argc + 2) * sizeof(char *));
+  c->argv[0] = pstrdup(c->pool, cmd->arg);
+  c->argv[1] = pcalloc(c->pool, sizeof(unsigned char));
+  *((unsigned char *) c->argv[1]) = eval_type;
 
-  argv = (char **) c->argv + 1;
+  argv = (char **) c->argv + 2;
 
   if (acl) {
     while (acl->nelts--) {
@@ -254,12 +273,32 @@ MODRET start_ifctxt(cmd_rec *cmd) {
   }
 
   *argv = NULL;
-
   return PR_HANDLED(cmd);
 }
 
 MODRET end_ifctxt(cmd_rec *cmd) {
   pr_parser_config_ctxt_close(NULL);
+
+  switch (ifsess_ctx) {
+    case IFSESS_CLASS_NUMBER:
+      if (strcasecmp("</IfClass>", cmd->argv[0]) == 0) {
+        ifsess_ctx = -1;
+      }
+      break;
+
+    case IFSESS_GROUP_NUMBER:
+      if (strcasecmp("</IfGroup>", cmd->argv[0]) == 0) {
+        ifsess_ctx = -1;
+      }
+      break;
+
+    case IFSESS_USER_NUMBER:
+      if (strcasecmp("</IfUser>", cmd->argv[0]) == 0) {
+        ifsess_ctx = -1;
+      }
+      break;
+  }
+
   return PR_HANDLED(cmd);
 }
 
@@ -299,38 +338,55 @@ MODRET ifsess_post_pass(cmd_rec *cmd) {
     if (list != NULL) {
       unsigned char mergein = FALSE;
 
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-      if (*((unsigned char *) list->argv[0]) == PR_EXPR_EVAL_REGEX) {
-        regex_t *preg = (regex_t *) list->argv[1];
+#ifdef PR_USE_REGEX
+      if (*((unsigned char *) list->argv[1]) == PR_EXPR_EVAL_REGEX) {
+        pr_regex_t *pre = list->argv[2];
 
-        if (session.group && regexec(preg, session.group, 0, NULL, 0) == 0)
-          mergein = TRUE;
+        if (session.group != NULL) {
+          pr_log_debug(DEBUG8, MOD_IFSESSION_VERSION
+            ": evaluating regexp pattern '%s' against subject '%s'",
+            pr_regexp_get_pattern(pre), session.group);
 
-        else if (session.groups) {
+          if (pr_regexp_exec(pre, session.group, 0, NULL, 0, 0, 0) == 0) {
+            mergein = TRUE;
+          }
+        }
+
+        if (mergein == FALSE &&
+            session.groups != NULL) {
           register int j = 0;
 
-          for (j = session.groups->nelts-1; j >= 0; j--)
-            if (regexec(preg, *(((char **) session.groups->elts) + j), 0,
-                NULL, 0) == 0)
+          for (j = session.groups->nelts-1; j >= 0; j--) {
+            char *suppl_group;
+
+            suppl_group = *(((char **) session.groups->elts) + j);
+
+            pr_log_debug(DEBUG8, MOD_IFSESSION_VERSION
+              ": evaluating regexp pattern '%s' against subject '%s'",
+              pr_regexp_get_pattern(pre), suppl_group);
+
+            if (pr_regexp_exec(pre, suppl_group, 0, NULL, 0, 0, 0) == 0) {
               mergein = TRUE;
+              break;
+            }
+          }
         }
+
       } else
-#endif /* HAVE_REGEX_H && HAVE_REGCOMP */
+#endif /* regex support */
     
-      if (*((unsigned char *) list->argv[0]) == PR_EXPR_EVAL_OR &&
-          pr_expr_eval_group_or((char **) &list->argv[1]) == TRUE)
+      if (*((unsigned char *) list->argv[1]) == PR_EXPR_EVAL_OR &&
+          pr_expr_eval_group_or((char **) &list->argv[2]) == TRUE) {
         mergein = TRUE;
 
-      else if (*((unsigned char *) list->argv[0]) == PR_EXPR_EVAL_AND &&
-          pr_expr_eval_group_and((char **) &list->argv[1]) == TRUE)
-        mergein = TRUE;
- 
-      if (pr_expr_eval_group_and((char **) &list->argv[0]) == TRUE)
+      } else if (*((unsigned char *) list->argv[1]) == PR_EXPR_EVAL_AND &&
+          pr_expr_eval_group_and((char **) &list->argv[2]) == TRUE) {
         mergein = TRUE;
+      }
 
       if (mergein) {
         pr_log_debug(DEBUG2, MOD_IFSESSION_VERSION
-          ": merging <IfGroup> directives in");
+          ": merging <IfGroup %s> directives in", (char *) list->argv[0]);
         ifsess_dup_set(session.pool, main_server->conf, c->subset);
 
         /* Add this config_rec pointer to the list of pointers to be
@@ -339,13 +395,20 @@ MODRET ifsess_post_pass(cmd_rec *cmd) {
         *((config_rec **) push_array(group_remove_list)) = c;
 
         resolve_deferred_dirs(main_server);
+
+        /* We need to call fixup_dirs() twice: once for any added <Directory>
+         * sections that use absolute paths, and again for any added <Directory>
+         * sections that use deferred-resolution paths (e.g. "~").
+         */
+        fixup_dirs(main_server, 0);
         fixup_dirs(main_server, CF_DEFER);
 
         ifsess_merged = TRUE;
 
-      } else
+      } else {
         pr_log_debug(DEBUG9, MOD_IFSESSION_VERSION
-          ": <IfGroup> not matched, skipping");
+          ": <IfGroup %s> not matched, skipping", (char *) list->argv[0]);
+      }
     }
 
     /* Note: it would be more efficient, memory-wise, to destroy the
@@ -377,27 +440,33 @@ MODRET ifsess_post_pass(cmd_rec *cmd) {
     if (list != NULL) {
       unsigned char mergein = FALSE;
 
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-      if (*((unsigned char *) list->argv[0]) == PR_EXPR_EVAL_REGEX) {
-        regex_t *preg = (regex_t *) list->argv[1];
+#ifdef PR_USE_REGEX
+      if (*((unsigned char *) list->argv[1]) == PR_EXPR_EVAL_REGEX) {
+        pr_regex_t *pre = list->argv[2];
 
-        if (regexec(preg, session.user, 0, NULL, 0) == 0)
+        pr_log_debug(DEBUG8, MOD_IFSESSION_VERSION
+          ": evaluating regexp pattern '%s' against subject '%s'",
+          pr_regexp_get_pattern(pre), session.user);
+
+        if (pr_regexp_exec(pre, session.user, 0, NULL, 0, 0, 0) == 0) {
           mergein = TRUE;
+        }
 
       } else
-#endif /* HAVE_REGEX_H && HAVE_REGCOMP */
+#endif /* regex support */
 
-      if (*((unsigned char *) list->argv[0]) == PR_EXPR_EVAL_OR &&
-          pr_expr_eval_user_or((char **) &list->argv[1]) == TRUE)
+      if (*((unsigned char *) list->argv[1]) == PR_EXPR_EVAL_OR &&
+          pr_expr_eval_user_or((char **) &list->argv[2]) == TRUE) {
         mergein = TRUE;
 
-      else if (*((unsigned char *) list->argv[0]) == PR_EXPR_EVAL_AND &&
-          pr_expr_eval_user_and((char **) &list->argv[1]) == TRUE)
+      } else if (*((unsigned char *) list->argv[1]) == PR_EXPR_EVAL_AND &&
+          pr_expr_eval_user_and((char **) &list->argv[2]) == TRUE) {
         mergein = TRUE;
+      }
 
       if (mergein) {
         pr_log_debug(DEBUG2, MOD_IFSESSION_VERSION
-          ": merging <IfUser> directives in");
+          ": merging <IfUser %s> directives in", (char *) list->argv[0]);
         ifsess_dup_set(session.pool, main_server->conf, c->subset);
 
         /* Add this config_rec pointer to the list of pointers to be
@@ -406,13 +475,20 @@ MODRET ifsess_post_pass(cmd_rec *cmd) {
         *((config_rec **) push_array(user_remove_list)) = c;
 
         resolve_deferred_dirs(main_server);
+
+        /* We need to call fixup_dirs() twice: once for any added <Directory>
+         * sections that use absolute paths, and again for any added <Directory>
+         * sections that use deferred-resolution paths (e.g. "~").
+         */
+        fixup_dirs(main_server, 0);
         fixup_dirs(main_server, CF_DEFER);
 
         ifsess_merged = TRUE;
 
-      } else
+      } else {
         pr_log_debug(DEBUG9, MOD_IFSESSION_VERSION
-          ": <IfUser> not matched, skipping");
+          ": <IfUser %s> not matched, skipping", (char *) list->argv[0]);
+      }
     }
 
     c = find_config_next(c, c->next, -1, IFSESS_USER_TEXT, FALSE);
@@ -429,18 +505,75 @@ MODRET ifsess_post_pass(cmd_rec *cmd) {
   if (ifsess_merged) {
     /* Try to honor any <Limit LOGIN> sections that may have been merged in. */
     if (!login_check_limits(TOPLEVEL_CONF, FALSE, TRUE, &found)) {
+      pr_log_debug(DEBUG3, MOD_IFSESSION_VERSION
+        ": %s %s: Limit access denies login",
+        session.anon_config ? "ANON" : C_USER, session.user);
+
       pr_log_auth(PR_LOG_NOTICE, "%s %s: Limit access denies login.",
         session.anon_config ? "ANON" : C_USER, session.user);
-      end_login(0);
+      pr_session_disconnect(&ifsession_module, PR_SESS_DISCONNECT_CONFIG_ACL,
+        "Denied by <Limit LOGIN>");
     }
   }
 
   return PR_DECLINED(cmd);
 }
 
+/* Event handlers
+ */
+
+#ifdef PR_SHARED_MODULE
+static void ifsess_mod_unload_ev(const void *event_data, void *user_data) {
+  if (strcmp("mod_ifsession.c", (const char *) event_data) == 0) {
+    pr_event_unregister(&ifsession_module, NULL, NULL);
+  }
+}
+#endif /* PR_SHARED_MODULE */
+
+static void ifsess_postparse_ev(const void *event_data, void *user_data) {
+  /* Make sure that all mod_ifsession sections have been properly closed. */
+
+  if (ifsess_ctx == -1) {
+    /* All sections properly closed; nothing to do. */
+    return;
+  }
+
+  switch (ifsess_ctx) {
+    case IFSESS_CLASS_NUMBER:
+      pr_log_pri(PR_LOG_ERR,
+        "error: unclosed <IfClass> context in config file");
+      break;
+
+    case IFSESS_GROUP_NUMBER:
+      pr_log_pri(PR_LOG_ERR,
+        "error: unclosed <IfGroup> context in config file");
+      break;
+
+    case IFSESS_USER_NUMBER:
+      pr_log_pri(PR_LOG_ERR,
+        "error: unclosed <IfUser> context in config file");
+      break;
+  }
+
+  pr_session_disconnect(&ifsession_module, PR_SESS_DISCONNECT_BAD_CONFIG, NULL);
+  return;
+}
+
 /* Initialization routines
  */
 
+static int ifsess_init(void) {
+#ifdef PR_SHARED_MODULE
+  pr_event_register(&ifsession_module, "core.module-unload",
+    ifsess_mod_unload_ev, NULL);
+#endif /* PR_SHARED_MODULE */
+
+  pr_event_register(&ifsession_module, "core.postparse",
+    ifsess_postparse_ev, NULL);
+
+  return 0;
+}
+
 static int ifsess_sess_init(void) {
   register unsigned int i = 0;
   config_rec *c = NULL;
@@ -459,28 +592,36 @@ static int ifsess_sess_init(void) {
     if (list != NULL) {
       unsigned char mergein = FALSE;
 
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-      if (*((unsigned char *) list->argv[0]) == PR_EXPR_EVAL_REGEX) {
-        regex_t *preg = (regex_t *) list->argv[1];
+#ifdef PR_USE_REGEX
+      if (*((unsigned char *) list->argv[1]) == PR_EXPR_EVAL_REGEX) {
+        pr_regex_t *pre = list->argv[2];
 
-        if (session.class && regexec(preg, session.class->cls_name, 0, NULL,
-            0) == 0)
-          mergein = TRUE;
+        if (session.class) {
+          pr_log_debug(DEBUG8, MOD_IFSESSION_VERSION
+            ": evaluating regexp pattern '%s' against subject '%s'",
+            pr_regexp_get_pattern(pre), session.class->cls_name);
+
+          if (pr_regexp_exec(pre, session.class->cls_name, 0, NULL, 0, 0,
+              0) == 0) {
+            mergein = TRUE;
+          }
+        }
 
       } else
-#endif /* HAVE_REGEX_H && HAVE_REGCOMP */
+#endif /* regex support */
 
-      if (*((unsigned char *) list->argv[0]) == PR_EXPR_EVAL_OR &&
-          pr_expr_eval_class_or((char **) &list->argv[1]) == TRUE)
+      if (*((unsigned char *) list->argv[1]) == PR_EXPR_EVAL_OR &&
+          pr_expr_eval_class_or((char **) &list->argv[2]) == TRUE) {
         mergein = TRUE;
 
-      else if (*((unsigned char *) list->argv[0]) == PR_EXPR_EVAL_AND &&
-          pr_expr_eval_class_and((char **) &list->argv[1]) == TRUE)
+      } else if (*((unsigned char *) list->argv[1]) == PR_EXPR_EVAL_AND &&
+          pr_expr_eval_class_and((char **) &list->argv[2]) == TRUE) {
         mergein = TRUE;
+      }
 
       if (mergein) {
         pr_log_debug(DEBUG2, MOD_IFSESSION_VERSION
-          ": merging <IfClass> directives in");
+          ": merging <IfClass %s> directives in", (char *) list->argv[0]);
         ifsess_dup_set(session.pool, main_server->conf, c->subset);
 
         /* Add this config_rec pointer to the list of pointers to be
@@ -488,13 +629,19 @@ static int ifsess_sess_init(void) {
          */
         *((config_rec **) push_array(class_remove_list)) = c;
 
+        /* We need to call fixup_dirs() twice: once for any added <Directory>
+         * sections that use absolute paths, and again for any added <Directory>
+         * sections that use deferred-resolution paths (e.g. "~").
+         */
+        fixup_dirs(main_server, 0);
         fixup_dirs(main_server, CF_DEFER);
 
         ifsess_merged = TRUE;
 
-      } else
+      } else {
         pr_log_debug(DEBUG9, MOD_IFSESSION_VERSION
-          ": <IfClass> not matched, skipping");
+          ": <IfClass %s> not matched, skipping", (char *) list->argv[0]);
+      }
     }
 
     c = find_config_next(c, c->next, -1, IFSESS_CLASS_TEXT, FALSE);
@@ -547,7 +694,7 @@ module ifsession_module = {
   NULL,
 
   /* Module initialization function */
-  NULL,
+  ifsess_init,
 
   /* Session initialization function */
   ifsess_sess_init,
